from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .core.potential_secret import PotentialSecret


class VerifiedResult(Enum):
    VERIFIED_FALSE = 1
    UNVERIFIED = 2
    VERIFIED_TRUE = 3

    @staticmethod
    def from_secret(secret: PotentialSecret) -> 'VerifiedResult':
        if secret.is_secret is None:
            return VerifiedResult.UNVERIFIED
        elif secret.is_secret:
            return VerifiedResult.VERIFIED_TRUE
        else:
            return VerifiedResult.VERIFIED_FALSE



"""
Responsible for input/output for audit related functions.
"""
import os
import platform
import sys
from enum import Enum

from ..custom_types import SecretContext
from ..util.color import AnsiColor
from ..util.color import colorize


def print_message(message: str) -> None:
    print(message)


def print_error(message: str) -> None:
    print(message, file=sys.stderr)


def clear_screen() -> None:     # pragma: no cover
    command = 'clear'
    if platform.system() == 'Windows':
        command = 'cls'
    os.system(command)


def print_context(context: SecretContext) -> None:
    if not context.snippet:     # pragma: no cover
        raise ValueError('You should be using `print_secret_not_found` instead.')

    _print_header(context)

    context.snippet.add_line_numbers()
    if context.secret.secret_value:
        context.snippet.highlight_line(context.secret.secret_value)
    else:
        context.snippet.target_line = colorize(context.snippet.target_line, AnsiColor.BOLD)
    print_message(str(context.snippet))

    print_message('-' * 10)


def print_secret_not_found(context: SecretContext) -> None:
    if context.snippet:     # pragma: no cover
        raise ValueError(
            'Are you sure you want to do this? The secret *was* found in this context. '
            'If you are certain you want to override this behavior, be sure to null out '
            'the `context.snippet` value.',
        )

    _print_header(context)

    print_message(str(context.error))
    print_message('-' * 10)


def _print_header(context: SecretContext) -> None:      # pragma: no cover
    print_message(
        '{secret} {current_count} {of} {total_count}'.format(
            secret=colorize('Secret:     ', AnsiColor.BOLD),
            current_count=colorize(str(context.current_index), AnsiColor.PURPLE),
            of=colorize('of', AnsiColor.BOLD),
            total_count=colorize(str(context.num_total_secrets), AnsiColor.PURPLE),
        ),
    )
    print_message(
        '{prefix} {filename}'.format(
            prefix=colorize('Filename:   ', AnsiColor.BOLD),
            filename=colorize(context.secret.filename, AnsiColor.PURPLE),
        ),
    )
    print_message(
        '{prefix} {secret_type}'.format(
            prefix=colorize('Secret Type:', AnsiColor.BOLD),
            secret_type=colorize(context.secret.type, AnsiColor.PURPLE),
        ),
    )
    print_message('-' * 10)

    if context.header:
        print(context.header)


def get_user_decision(
    prompt_secret_decision: bool = True,
    can_step_back: bool = False,
) -> 'InputOptions':
    """
    :param prompt_secret_decision: if False, won't ask to label secret.
        e.g. if the secret isn't found on the line
    """
    prompter = UserPrompt(allow_labelling=prompt_secret_decision, allow_backstep=can_step_back)

    user_input = None
    while user_input not in prompter.valid_input:
        if user_input:
            print('Invalid input.')     # type: ignore # Statement unreachable? Come on mypy...

        user_input = input(str(prompter))
        if user_input:
            user_input = user_input[0].upper()

    return InputOptions(user_input)


class InputOptions(Enum):
    YES = 'Y'
    NO = 'N'
    SKIP = 'S'
    BACK = 'B'
    QUIT = 'Q'


class UserPrompt:
    def __init__(self, allow_labelling: bool, allow_backstep: bool) -> None:
        options = []
        if allow_labelling:
            options += [InputOptions.YES, InputOptions.NO]

        options.append(InputOptions.SKIP)
        if allow_backstep:
            options.append(InputOptions.BACK)

        options.append(InputOptions.QUIT)

        self.valid_input = {option.name[0] for option in options}
        self.options = [option.name.lower() for option in options]

    def __str__(self) -> str:
        if 'Y' in self.valid_input:
            output = 'Is this a secret that should be committed to this repository?'
        else:
            output = 'What would you like to do?'

        options = ', '.join([f'({option[0]}){option[1:]}' for option in self.options])

        return output + ' ' + options + ': '



"""
This plugin finds JWT tokens
"""
import base64
import json
import re
from typing import Generator

from .base import RegexBasedDetector


class JwtTokenDetector(RegexBasedDetector):
    """Scans for JWTs."""
    secret_type = 'JSON Web Token'
    denylist = [
        re.compile(r'eyJ[A-Za-z0-9-_=]+\.[A-Za-z0-9-_=]+\.?[A-Za-z0-9-_.+/=]*?'),
    ]

    def analyze_string(self, string: str) -> Generator[str, None, None]:
        yield from filter(
            self.is_formally_valid,
            super().analyze_string(string),
        )

    @staticmethod
    def is_formally_valid(token: str) -> bool:
        parts = token.split('.')
        for idx, part_str in enumerate(parts):
            try:
                part = part_str.encode('ascii')
                # https://github.com/magical/jwt-python/blob/2fd976b41111031313107792b40d5cfd1a8baf90/jwt.py#L49
                # https://github.com/jpadilla/pyjwt/blob/3d47b0ea9e5d489f9c90ee6dde9e3d9d69244e3a/jwt/utils.py#L33
                m = len(part) % 4
                if m == 1:
                    raise TypeError('Incorrect padding')
                elif m == 2:
                    part += '=='.encode('utf-8')
                elif m == 3:
                    part += '==='.encode('utf-8')
                b64_decoded = base64.urlsafe_b64decode(part)
                if idx < 2:
                    _ = json.loads(b64_decoded.decode('utf-8'))
            except (TypeError, ValueError, UnicodeDecodeError):
                return False

        return True



"""
NOTE(2020-11-07|domanchi): We probably can use `python-semver` to do this. However, at the
time of writing, it doesn't look like it's ready for production. Therefore, this implements
a very basic version of it.
"""
from typing import Any


class Version:
    def __init__(self, version: str) -> None:
        self.major, self.minor, self.patch = map(int, version.split('.'))

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, Version):
            raise NotImplementedError

        if self.major < other.major:
            return True
        if self.major > other.major:
            return False

        if self.minor < other.minor:
            return True
        if self.minor > other.minor:
            return False

        if self.patch < other.patch:
            return True

        return False

    def __gt__(self, other: Any) -> bool:
        return not self.__lt__(other)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Version):
            raise NotImplementedError

        return (
            self.major == other.major
            and self.minor == other.minor
            and self.patch == other.patch
        )

    def __le__(self, other: Any) -> bool:
        return self.__lt__(other) or self.__eq__(other)

    def __ge__(self, other: Any) -> bool:
        return self.__gt__(other) or self.__eq__(other)



"""
This plugin searches for NPM tokens
"""
import re

from detect_secrets.plugins.base import RegexBasedDetector


class NpmDetector(RegexBasedDetector):
    """Scans for NPM tokens."""
    secret_type = 'NPM tokens'

    denylist = [
        # npmrc authToken
        # ref. https://stackoverflow.com/questions/53099434/using-auth-tokens-in-npmrc
        re.compile(r'\/\/.+\/:_authToken=\s*((npm_.+)|([A-Fa-f0-9-]{36})).*'),
    ]



from typing import Union

import requests

from ..constants import VerifiedResult
from .base import RegexBasedDetector


class IbmCloudIamDetector(RegexBasedDetector):
    """Scans for IBM Cloud IAM Key."""

    secret_type = 'IBM Cloud IAM Key'

    # opt means optional
    opt_ibm_cloud_iam = r'(?:ibm(?:_|-|)cloud(?:_|-|)iam|cloud(?:_|-|)iam|' + \
        r'ibm(?:_|-|)cloud|ibm(?:_|-|)iam|ibm|iam|cloud|)'
    opt_dash_underscore = r'(?:_|-|)'
    opt_api = r'(?:api|)'
    key_or_pass = r'(?:key|pwd|password|pass|token)'
    secret = r'([a-zA-Z0-9_\-]{44}(?![a-zA-Z0-9_\-]))'
    denylist = [
        RegexBasedDetector.build_assignment_regex(
            prefix_regex=opt_ibm_cloud_iam + opt_dash_underscore + opt_api,
            secret_keyword_regex=key_or_pass,
            secret_regex=secret,
        ),
    ]

    def verify(self, secret: str) -> VerifiedResult:
        response = verify_cloud_iam_api_key(secret)

        return VerifiedResult.VERIFIED_TRUE if response.status_code == 200 \
            else VerifiedResult.VERIFIED_FALSE


def verify_cloud_iam_api_key(apikey: Union[str, bytes]) -> requests.Response:  # pragma: no cover
    if type(apikey) == bytes:
        apikey = apikey.decode('UTF-8')

    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Accept': 'application/json',
    }
    response = requests.post(
        'https://iam.cloud.ibm.com/identity/token',
        headers=headers,
        data={
            'grant_type': 'urn:ibm:params:oauth:grant-type:apikey',
            'apikey': apikey,
        },
    )
    return response



import re
from typing import List

import requests

from ..constants import VerifiedResult
from ..util.code_snippet import CodeSnippet
from .base import RegexBasedDetector


class CloudantDetector(RegexBasedDetector):
    """Scans for Cloudant credentials."""

    secret_type = 'Cloudant Credentials'

    # opt means optional
    dot = r'\.'
    cl_account = r'[\w\-]+'
    cl = r'(?:cloudant|cl|clou)'
    opt_api = r'(?:api|)'
    cl_key_or_pass = opt_api + r'(?:key|pwd|pw|password|pass|token)'
    cl_pw = r'([0-9a-f]{64})'
    cl_api_key = r'([a-z]{24})'
    colon = r'\:'
    at = r'\@'
    http = r'(?:https?\:\/\/)'
    cloudant_api_url = r'cloudant\.com'
    denylist = [
        RegexBasedDetector.build_assignment_regex(
            prefix_regex=cl,
            secret_keyword_regex=cl_key_or_pass,
            secret_regex=cl_pw,
        ),
        RegexBasedDetector.build_assignment_regex(
            prefix_regex=cl,
            secret_keyword_regex=cl_key_or_pass,
            secret_regex=cl_api_key,
        ),
        re.compile(
            r'{http}{cl_account}{colon}{cl_pw}{at}{cl_account}{dot}{cloudant_api_url}'.format(
                http=http,
                colon=colon,
                cl_account=cl_account,
                cl_pw=cl_pw,
                at=at,
                dot=dot,
                cloudant_api_url=cloudant_api_url,
            ),
            flags=re.IGNORECASE,
        ),
        re.compile(
            r'{http}{cl_account}{colon}{cl_api_key}{at}{cl_account}{dot}{cloudant_api_url}'.format(
                http=http,
                colon=colon,
                cl_account=cl_account,
                cl_api_key=cl_api_key,
                at=at,
                dot=dot,
                cloudant_api_url=cloudant_api_url,
            ),
            flags=re.IGNORECASE,
        ),
    ]

    def verify(       # type: ignore[override]  # noqa: F821
        self,
        secret: str,
        context: CodeSnippet,
    ) -> VerifiedResult:
        hosts = find_account(context)
        if not hosts:
            return VerifiedResult.UNVERIFIED

        for host in hosts:
            return verify_cloudant_key(host, secret)

        return VerifiedResult.VERIFIED_FALSE


def find_account(context: CodeSnippet) -> List[str]:
    opt_hostname_keyword = r'(?:hostname|host|username|id|user|userid|user-id|user-name|' \
        'name|user_id|user_name|uname|account)'
    account = r'(\w[\w\-]*)'
    opt_basic_auth = r'(?:[\w\-:%]*\@)?'

    regexes = (
        RegexBasedDetector.build_assignment_regex(
            prefix_regex=CloudantDetector.cl,
            secret_keyword_regex=opt_hostname_keyword,
            secret_regex=account,
        ),
        re.compile(
            r'{http}{opt_basic_auth}{cl_account}{dot}{cloudant_api_url}'.format(
                http=CloudantDetector.http,
                opt_basic_auth=opt_basic_auth,
                cl_account=account,
                dot=CloudantDetector.dot,
                cloudant_api_url=CloudantDetector.cloudant_api_url,
            ),
            flags=re.IGNORECASE,
        ),
    )

    return [
        match
        for line in context
        for regex in regexes
        for match in regex.findall(line)
    ]


def verify_cloudant_key(hostname: str, token: str) -> VerifiedResult:
    headers = {'Content-type': 'application/json'}
    request_url = 'https://{hostname}:' \
        '{token}' \
        '@{hostname}.' \
        'cloudant.com'.format(
            hostname=hostname,
            token=token,
        )

    try:
        response = requests.get(
            request_url,
            headers=headers,
        )
    except requests.exceptions.RequestException:
        return VerifiedResult.UNVERIFIED

    if response.status_code == 200:
        return VerifiedResult.VERIFIED_TRUE
    else:
        return VerifiedResult.VERIFIED_FALSE



import re
from typing import List

import requests

from ..constants import VerifiedResult
from ..util.code_snippet import CodeSnippet
from .base import RegexBasedDetector


class SoftlayerDetector(RegexBasedDetector):
    """Scans for Softlayer credentials."""

    secret_type = 'SoftLayer Credentials'

    # opt means optional
    sl = r'(?:softlayer|sl)(?:_|-|)(?:api|)'
    key_or_pass = r'(?:key|pwd|password|pass|token)'
    secret = r'([a-z0-9]{64})'
    denylist = [
        RegexBasedDetector.build_assignment_regex(
            prefix_regex=sl,
            secret_keyword_regex=key_or_pass,
            secret_regex=secret,
        ),

        re.compile(
            r'(?:http|https)://api.softlayer.com/soap/(?:v3|v3.1)/([a-z0-9]{64})',
            flags=re.IGNORECASE,
        ),
    ]

    def verify(       # type: ignore[override]  # noqa: F821
        self,
        secret: str,
        context: CodeSnippet,
    ) -> VerifiedResult:
        usernames = find_username(context)
        if not usernames:
            return VerifiedResult.UNVERIFIED

        for username in usernames:
            return verify_softlayer_key(username, secret)

        return VerifiedResult.VERIFIED_FALSE


def find_username(context: CodeSnippet) -> List[str]:
    # opt means optional
    username_keyword = (
        r'(?:'
        r'username|id|user|userid|user-id|user-name|'
        r'name|user_id|user_name|uname'
        r')'
    )
    username = r'(\w(?:\w|_|@|\.|-)+)'
    regex = re.compile(
        RegexBasedDetector.build_assignment_regex(
            prefix_regex=SoftlayerDetector.sl,
            secret_keyword_regex=username_keyword,
            secret_regex=username,
        ),
    )

    return [
        match
        for line in context
        for match in regex.findall(line)
    ]


def verify_softlayer_key(username: str, token: str) -> VerifiedResult:
    headers = {'Content-type': 'application/json'}
    try:
        response = requests.get(
            'https://api.softlayer.com/rest/v3/SoftLayer_Account.json',
            auth=(username, token), headers=headers,
        )
    except requests.exceptions.RequestException:
        return VerifiedResult.UNVERIFIED

    if response.status_code == 200:
        return VerifiedResult.VERIFIED_TRUE
    else:
        return VerifiedResult.VERIFIED_FALSE



import importlib.util
import os
import pkgutil
from importlib import import_module
from types import ModuleType
from typing import Any
from typing import Callable
from typing import Iterable
from typing import List
from typing import Optional
from typing import Type

from ..exceptions import InvalidFile


def import_types_from_package(
    root: ModuleType,
    filter: Callable[[Any], bool],
) -> Iterable[Type]:
    output: List[Type] = []
    modules = get_modules_from_package(root)

    for module_path in modules:
        module = import_module(module_path)
        output.extend(import_types_from_module(module, filter))

    return output


def import_types_from_module(
    module: ModuleType,
    filter: Callable[[Any], bool],
) -> Iterable[Type]:
    output = []
    for name in dir(module):
        if name.startswith('_'):
            continue

        attribute = getattr(module, name)
        if filter(attribute):
            continue

        output.append(attribute)

    return output


def import_modules_from_package(
    root: ModuleType,
    filter: Callable[[str], bool],
) -> Iterable[ModuleType]:
    output = []
    modules = get_modules_from_package(root)

    # NOTE: It should be auto-sorted, but let's just do it for sanity sake.
    # This sorting is required for performing upgrades in order.
    for module_path in sorted(modules):
        if filter(module_path):
            continue

        output.append(import_module(module_path))

    return output


def import_file_as_module(filename: str, name: Optional[str] = None) -> ModuleType:
    """
    NOTE(2020-11-09|domanchi): We're essentially executing arbitrary code here, so some thoughts
    should be recorded as to the security of this feature. This should not add any additional
    security risk, given the following assumptions hold true:

      1. detect-secrets is not used in an environment that has privileged access (more
         than the current user), OR
      2. detect-secrets (when running in a privileged context) does not accept arbitrary
         user input that feeds into this function (e.g. custom plugins).

    The first assumption should be rather self-explanatory: if you are running detect-secrets
    in a context that has the same permissions as you, you can import any code you want, since
    this acts more of a utility function than a security flaw. If you're going to do it *anyway*,
    let's just make your life easier.

    The second assumption should also be pretty straight-forward: don't trust user input,
    especially if it's going to be executed as that privileged user, unless you want a privilege
    escalation vulnerability. detect-secrets is not going to do any sanitization of user input
    for you.
    """
    if not os.path.exists(filename):
        raise FileNotFoundError

    if not name:
        # NOTE: After several trial and error attempts, I could not discern the importance
        # of this field, in this context. Hence, I don't think it matters that much.
        name = os.path.splitext(os.path.basename(filename))[0]

    # Source: https://stackoverflow.com/a/67692/13340678
    spec = importlib.util.spec_from_file_location(name, filename)
    if not spec:
        raise InvalidFile

    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)                 # type: ignore
    module.__path__ = os.path.abspath(filename)     # type: ignore

    return module


def get_modules_from_package(root: ModuleType) -> Iterable[str]:
    return [
        module
        for _, module, is_package in pkgutil.walk_packages(
            root.__path__, prefix=f'{root.__name__}.',
        )
        if not is_package
    ]



from typing import Generator
from typing import List

from .color import AnsiColor
from .color import colorize
from detect_secrets.exceptions import SecretNotFoundOnSpecifiedLineError


def get_code_snippet(
    lines: List[str],
    line_number: int,
    lines_of_context: int = 5,
) -> 'CodeSnippet':
    """
    :param lines: an iterator of lines in the file
    :param line_number: line which you want to focus on
    :param lines_of_context: how many lines to display around the line you want
        to focus on.
    """
    target_line_index = line_number - 1
    end_line_index = target_line_index + lines_of_context + 1

    if target_line_index <= lines_of_context:
        start_line_index = 0
    else:
        start_line_index = target_line_index - lines_of_context
        target_line_index = lines_of_context

    return CodeSnippet(
        snippet=lines[start_line_index:end_line_index],
        start_line=start_line_index,
        target_index=target_line_index,
    )


class CodeSnippet:

    def __init__(self, snippet: List[str], start_line: int, target_index: int) -> None:
        """
        :param snippet: lines of code extracted from file
        :param start_line: first line number in segment
        :param target_index: index in snippet of target line
        """
        self.lines = snippet
        self.start_line = start_line
        self.target_index = target_index

    @property
    def target_line(self) -> str:
        return self.lines[self.target_index]

    @target_line.setter
    def target_line(self, value: str) -> None:
        self.lines[self.target_index] = value

    @property
    def previous_line(self) -> str:
        if self.target_index == 0 or len(self.lines) < self.target_index:
            return ''
        return self.lines[self.target_index - 1]

    def add_line_numbers(self) -> 'CodeSnippet':
        for index, line in enumerate(self.lines):
            self.lines[index] = u'{}:{}'.format(
                self.get_line_number(self.start_line + index + 1),
                line,
            )

        return self

    def highlight_line(self, payload: str) -> 'CodeSnippet':
        """
        :param payload: string to highlight, on chosen line
        """
        try:
            index_of_payload = self.target_line.lower().index(payload.lower())
            end_of_payload = index_of_payload + len(payload)

            self.target_line = u'{}{}{}'.format(
                self.target_line[:index_of_payload],
                self.apply_highlight(self.target_line[index_of_payload:end_of_payload]),
                self.target_line[end_of_payload:],
            )

            return self
        except ValueError:
            raise SecretNotFoundOnSpecifiedLineError(self.target_index)

    def get_line_number(self, line_number: int) -> str:
        """Broken out, for custom colorization."""
        return colorize(str(line_number), AnsiColor.LIGHT_GREEN)

    def apply_highlight(self, payload: str) -> str:
        """Broken out, for custom colorization."""
        return colorize(payload, AnsiColor.RED_BACKGROUND)

    def __str__(self) -> str:
        return '\n'.join(self.lines)

    def __iter__(self) -> Generator[str, None, None]:
        yield from self.lines



from typing import Any
from typing import Dict


def upgrade(baseline: Dict[str, Any]) -> None:
    if 'exclude_regex' in baseline:
        baseline['exclude'] = {
            'files': baseline.pop('exclude_regex'),
            'lines': None,
        }

    baseline['word_list'] = {
        'file': None,
        'hash': None,
    }



import os
import re
import string
from functools import lru_cache
from typing import Optional
from typing import Pattern

from detect_secrets.plugins.base import BasePlugin
from detect_secrets.plugins.base import RegexBasedDetector


def is_sequential_string(secret: str) -> bool:
    sequences = (
        # Base64 letters first
        (
            string.ascii_uppercase +
            string.ascii_uppercase +
            string.digits +
            '+/'
        ),

        # Base64 numbers first
        (
            string.digits +
            string.ascii_uppercase +
            string.ascii_uppercase +
            '+/'
        ),

        # We don't have a specific sequence for alphabetical
        # sequences, since those will happen to be caught by the
        # base64 checks.

        # Alphanumeric sequences
        (string.digits + string.ascii_uppercase) * 2,

        # Capturing any number sequences
        string.digits * 2,

        string.hexdigits.upper() + string.hexdigits.upper(),
        string.ascii_uppercase + '=/',
    )

    uppercase = secret.upper()
    for sequential_string in sequences:
        if uppercase in sequential_string:
            return True

    return False


def is_potential_uuid(secret: str) -> bool:
    return bool(_get_uuid_regex().search(secret))


@lru_cache(maxsize=1)
def _get_uuid_regex() -> Pattern:
    return re.compile(
        r'[a-f0-9]{8}\-[a-f0-9]{4}\-[a-f0-9]{4}\-[a-f0-9]{4}\-[a-f0-9]{12}',
        re.IGNORECASE,
    )


def is_likely_id_string(secret: str, line: str, plugin: Optional[BasePlugin] = None) -> bool:
    try:
        index = line.index(secret)
    except ValueError:
        return False

    return (not plugin or not isinstance(plugin, RegexBasedDetector)) \
        and bool(_get_id_detector_regex().search(line, pos=0, endpos=index))


@lru_cache(maxsize=1)
def _get_id_detector_regex() -> Pattern:
    """
    Regex Details:
    ^(id|myid|userid) -> Common id identifiers with no prefix
    _id               -> id identifier with prefixes allowed
    s?                -> Optional plural id identifier
    [^a-z0-9]         -> Non-letter/numeric character
    """
    return re.compile(r'(^(id|myid|userid)|_id)s?[^a-z0-9]', re.IGNORECASE)


def is_non_text_file(filename: str) -> bool:
    _, ext = os.path.splitext(filename)
    return ext in IGNORED_FILE_EXTENSIONS


# We don't scan files with these extensions.
# Note: We might be able to do this better with
#       `subprocess.check_output(['file', filename])`
#       and look for "ASCII text", but that might be more expensive.
#
#       Definitely something to look into, if this list gets unruly long.
IGNORED_FILE_EXTENSIONS = set(
    (
        '.7z',
        '.bin',
        '.bmp',
        '.bz2',
        '.class',
        '.css',
        '.dmg',
        '.doc',
        '.eot',
        '.exe',
        '.gif',
        '.gz',
        '.ico',
        '.iml',
        '.ipr',
        '.iws',
        '.jar',
        '.jpg',
        '.jpeg',
        '.lock',
        '.map',
        '.mo',
        '.pdf',
        '.png',
        '.prefs',
        '.psd',
        '.rar',
        '.realm',
        '.s7z',
        '.sum',
        '.svg',
        '.tar',
        '.tif',
        '.tiff',
        '.ttf',
        '.webp',
        '.woff',
        '.xls',
        '.xlsx',
        '.zip',
    ),
)


def is_templated_secret(secret: str) -> bool:
    """
    Filters secrets that are shaped like: {secret}, <secret>, or ${secret}.
    """
    try:
        if (
            (secret[0] == '{' and secret[-1] == '}')
            or (secret[0] == '<' and secret[-1] == '>')
            or (secret[0] == '$' and secret[1] == '{' and secret[-1] == '}')
        ):
            return True
    except IndexError:
        # Any one character secret (that causes this to raise an IndexError) is highly
        # likely to be a false positive (or if a true positive, INCREDIBLY weak password).
        return True

    return False


def is_prefixed_with_dollar_sign(secret: str) -> bool:
    # NOTE: This is broken out into its own function since it has more chance of increasing
    # false negatives than `is_templated_secret` (e.g. secrets that actually start with a $).
    # This is best used with files that actually use this as a means of referencing variables.
    # TODO: More intelligent filetype handling?
    return secret[0] == '$'


def is_indirect_reference(line: str) -> bool:
    """
    Filters secrets that take the form of:

        secret = get_secret_key()

    or

        secret = request.headers['apikey']
    """
    # Constrain line length as the heuristic's intention is to target lines that resemble
    # function calls. The constraint avoids catastrophic backtracking failures of the regex.
    if len(line) > 1000:
        return False
    return bool(_get_indirect_reference_regex().search(line))


@lru_cache(maxsize=1)
def _get_indirect_reference_regex() -> Pattern:
    # Regex details:
    #   ([^\v=!:]*)     ->  Something before the assignment or comparison
    #   \s*             ->  Some optional whitespaces
    #   (:=?|[!=]{1,3}) ->  Assignment or comparison: :=, =, ==, ===, !=, !==
    #   \s*             ->  Some optional whitespaces
    #   (
    #       [\w.-]+     ->  Some alphanumeric character, dot or -
    #       [\[\(]      ->  Start of indirect reference: [ or (
    #       [^\v]*      ->  Something except line breaks
    #       [\]\)]      ->  End of indirect reference: ] or )
    #   )
    return re.compile(r'([^\v=!:]*)\s*(:=?|[!=]{1,3})\s*([\w.-]+[\[\(][^\v]*[\]\)])')


def is_lock_file(filename: str) -> bool:
    return os.path.basename(filename) in {
        'Brewfile.lock.json',
        'Cartfile.resolved',
        'composer.lock',
        'Gemfile.lock',
        'Package.resolved',
        'package-lock.json',
        'Podfile.lock',
        'yarn.lock',
        'Pipfile.lock',
        'poetry.lock',
        'Cargo.lock',
        'packages.lock.json',
    }


def is_not_alphanumeric_string(secret: str) -> bool:
    """
    This assumes that secrets should have at least ONE letter in them.
    This helps avoid clear false positives, like `*****`.
    """
    return not bool(set(string.ascii_letters) & set(secret))


def is_swagger_file(filename: str) -> bool:
    """
    Filters swagger files and paths, like swagger-ui.html or /swagger/.
    """
    return bool(_get_swagger_regex().search(filename))


@lru_cache(maxsize=1)
def _get_swagger_regex() -> Pattern:
    return re.compile(r'.*swagger.*')



"""
This plugin searches for GitHub tokens
"""
import re

from detect_secrets.plugins.base import RegexBasedDetector


class GitHubTokenDetector(RegexBasedDetector):
    """Scans for GitHub tokens."""
    secret_type = 'GitHub Token'

    denylist = [
        # ref. https://github.blog/2021-04-05-behind-githubs-new-authentication-token-formats/
        re.compile(r'(ghp|gho|ghu|ghs|ghr)_[A-Za-z0-9_]{36}'),
    ]
